
{ The  Emulator (C) 1989-91 Todor Todorov
  Debugging routines and IRQ support, init code

  29 Sep 1991
  16 Oct 1991 }

{$A-,B-,D+,E-,F-,I-,L+,N-,O-,R-,S-,V-}
{$M 1024,0,0}

program pyldin;
  uses crt, dos;
  const trap = $60; { Trap interrupt }
	brk = $3e; { Breakpoint instruction }
	sysdate = $18; { Some fixed addresses in memory }
	piadra = $e628;
	piacra = $e62a;
	piacrb = $e62b;
	rompage = $e6f0;
	resmagic = $ed00;
	synchroadr = $ed04;
	kbdstore = $ed0c;
	kbdload = $ed0e;
	oldkey = $ed31;
	intpags = $ed80;
	intadrs = $ee00;
	modelbyte = $fff7;
	irqvect = $fff8;
	resvect = $fffe;
  var oldusr, oldkbd, oldtim: procedure; { Saved interrupt vectors }
      scrbuf: array[1..8192] of byte; { Area to save text screen - must be at
					a paragraph boundary!!! This would
					be OK with Turbo Pascal 5.0 only }
      crtregs: array[0..2] of word;   { Area to save crt regusters R10..R15 -
					must be immediately after scrbuf!!! }
      oldc: word; { Saved cursor shape }
      regs: registers; { For calling interrupts }
      mainseg: word; { Emulator address space segment }
      ramseg: word; { Emulator page RAM segment }
      romseg: word; { Emulator page ROM segment }
      oldlvl: word; { Saved stack pointer when end in progress }
      model: boolean; { True if 601A }
      oldpage: byte; { Current page }
      newpage: word; { Debugger page }
      patchofs: array[0..399] of word; { Array of patch offsets in emu segment }
      patchtop: word; { Length of patch offsets array }
      patchrts: word; { Patch for rts instruction }
      patchrti: word; { Patch for rti instruction }
      r: record { 6800 registers }
	   a, b, x, sp, pc: word
	 end;
      flags: array[0..7] of boolean;
      bpt: array[0..6] of { Breakpoints array }
	record
	  adr: word; { Breakpoint address }
	  page: byte; { Breakpoint page, 0 for any }
	  old: byte; { Saved byte }
	  count, occur: word;
	  bad: boolean; { Code changed }
	end;
      wpt: array[0..6] of word; { Watchpoints array }
      col: array[1..7] of word; { Windows colors }
      line: string[16]; { Line buffer }
      loadpath: dirstr; { Directory that has the emulator files }
      attr, textaddr: word; xpos, ypos: byte; { Parameters for write routines }
      keyflags: byte absolute $40:$17;
  const debugnow: boolean = false; { True if debugging }
	trapnow: boolean = false; { True if trap in progress }
	endnow: boolean = false; { True if end in progress }
	irqnow: boolean = false; { True if interrupt request is pending }
	hotkey: byte = 0; { Hotkey scan code, 0 if none }
	keytim: byte = 0; { Time left for the key to last }
	skipbpt: byte = $ff; { Number of the breakpoint currently skipped }
	timecnt: byte = 0; { Timer counter }
	emuseg: word = 0; { Emulator code segment }
	cursadr: word = 0; { Code window cursor address }
	codadr: word = 0; { Code window start address }
	datadr: word = 0; { Data window start address }
	stkofs: integer = 0; { Stack window start offset }
	posx: array[1..7] of byte = { Windows cursor x positions }
	(4,4,44,58,58,4,62);
	posy: array[1..7] of byte = { Windows cursor y positions }
	(3,5,13,2,9,16,16);
	window: byte = 2; { Current Window }
	echo: boolean = false; { True if writec outputs also to line buffer }
	bin: boolean = false; { True if binary input }
	interrupts: boolean = false; { True if /E interrupt enable }
  const xlat: array[$3b..$8a] of byte = { Keyboard translate table }
  ($c9,$ca,$cb,$cc,$cd,$ce,$cf,$d0,$d1,$d2,0,0,$c5,$c4,$c8,0,
   $c1,0,$c2,0,$c6,$c3,$c7,$fa,$7f,$d5,$d6,$d7,$d8,$d9,$da,$db,
   $dc,$dd,$de,$e1,$e2,$e3,$e4,$e5,$e6,$e7,$e8,$e9,$ea,0,0,0,
   0,0,0,0,0,0,0,0,$f1,$f2,$f5,$f7,$f6,0,0,0,
   0,0,0,0,0,0,0,0,0,$f8,$d3,$d4,$df,$e0,$eb,$ec);

  { Routines to write to screen }

  procedure setxy(x, y: byte);
  begin
    textaddr := (y-1)*160+(x-1) shl 1;
    xpos := x; ypos := y
  end;
  procedure writex(c: char);
  begin
    mem[$b800:textaddr] := ord(c)
  end;
  procedure writec(c: char);
  begin
    if echo then line := line+c;
    memw[$b800:textaddr] := attr+ord(c);
    inc(textaddr,2);
    inc(xpos)
  end;
  procedure writes(s: string);
    var i: byte;
  begin
    for i := 1 to length(s) do
      if s[i] = #8
	then dec(textaddr,2)
	else writec(s[i])
  end;

  { Routines to output hex numbers }

  function cvt(c: byte): byte;
    inline($58/ 			     { POP  AX	  }
	   $3c/$a/			     { CMP  AL,0A }
	   $f5/ 			     { CMC	  }
	   $14/$30/			     { ADC  AL,30 }
	   $27);			     { DAA	  }
  procedure hexout(h: byte);
  begin
    writec(chr(cvt(h)))
  end;
  procedure bytout(b: byte);
  begin
    writec(chr(cvt(b shr 4)));
    writec(chr(cvt(b and $f)))
  end;
  procedure wrdout(w: word);
  begin
    bytout(hi(w));
    bytout(lo(w))
  end;
  procedure wrdcol(w: word);
  begin
    wrdout(w);
    writes(':  ')
  end;

  { Routines to draw boxes }

  procedure putc(c: char; x, y: byte);
  begin
    setxy(x,y);
    writec(c)
  end;
  procedure fillx(i, x, y: byte);
  begin
    setxy(x,y);
    for i := 1 to i do writec(#$c4)
  end;
  procedure filly(i, x, y: byte);
  begin
    for i := 0 to i-1 do putc(#$d3,x,y+i)
  end;

  { Routines to access data in memory }

  function getp(adr: word; page: byte): byte;
    var ofs: word;
  begin
    ofs := adr-$c000;
    if ofs >= $2000
      then getp := mem[mainseg:adr]
      else
	if page < 8
	  then getp := mem[ramseg:ofs]
	  else getp := mem[romseg:page shl 13+ofs]
  end;
  procedure putp(adr: word; dat, page: byte);
    var ofs: word;
  begin
    ofs := adr-$c000;
    if ofs >= $2000
      then mem[mainseg:adr] := dat
      else
	if page < 8
	  then mem[ramseg:ofs] := dat
	  else mem[romseg:page shl 13+ofs] := dat
  end;
  function getw(adr: word): word;
    var ofs: word;
  begin
    getw := swap(getp(adr,newpage))+getp(adr+1,newpage)
  end;
  procedure putwp(adr: word; dat: word; page: byte);
    var ofs: word;
  begin
    putp(adr,hi(dat),page);
    putp(adr+1,lo(dat),page)
  end;
  procedure putw(adr: word; dat: word);
  begin
    putwp(adr, dat, newpage)
  end;
  function get(adr: word): byte;
  begin
    get := getp(adr,newpage)
  end;
  procedure put(adr: word; dat: byte);
  begin
    putp(adr,dat,newpage)
  end;

  { Routines to move through code }

  const implied = 0;
	relative = 1;
	immed = 2;
	direct = 3;
	indexed = 4;
	extended = 5;
  var opcod, opmod, len: byte;
  procedure getmod;
    const mode: array[byte] of 0..5 =
	 (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	  1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,3,
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	  4,0,0,4,4,0,4,4,4,4,4,0,4,4,4,4,
	  5,0,0,5,5,0,5,5,5,5,5,0,5,5,5,5,
	  2,2,2,0,2,2,2,0,2,2,2,2,2,1,2,0,
	  3,3,3,0,3,3,3,3,3,3,3,3,3,0,3,3,
	  4,4,4,0,4,4,4,4,4,4,4,4,4,4,4,4,
	  5,5,5,0,5,5,5,5,5,5,5,5,5,5,5,5,
	  2,2,2,0,2,2,2,0,2,2,2,2,0,0,2,0,
	  3,3,3,0,3,3,3,3,3,3,3,3,0,0,3,3,
	  4,4,4,0,4,4,4,4,4,4,4,4,0,0,4,4,
	  5,5,5,0,5,5,5,5,5,5,5,5,0,0,5,5);
	  modelen: array[0..5] of 1..3 = (1,2,2,2,2,3);
  begin
    opmod := mode[opcod];
    len := modelen[opmod];
    case opcod of
      $8c, $8e, $ce: inc(len)
    end
  end;
  procedure getlen(w: word);
  begin
    opcod := get(w);
    getmod
  end;
  procedure up(n: byte; reset: boolean);
    const maxbak = 32;
    var i, j, k: word;
	stat: array[1..3] of word;
  begin
    for n := 1 to n do
      begin
	stat[1] := 0; stat[2] := 0; stat[3] := 0;
	for k := 1 to maxbak do
	  begin
	    i := codadr-k;
	    repeat
	      j := i;
	      getlen(i);
	      inc(i,len)
	    until i >= codadr;
	    if i = codadr then inc(stat[codadr-j])
	  end;
	i := stat[1];
	if stat[2] > stat[3]
	  then
	    begin
	      if stat[2] > i then
		begin
		  i := stat[2];
		  if i > 0 then dec(codadr)
		end
	    end
	  else
	    if stat[3] > i then
	      begin
		i := stat[3];
		if i > 0 then dec(codadr,2)
	      end;
	if (i > 0) or not reset
	  then dec(codadr)
	  else posy[2] := 5
      end
  end;
  procedure down(n: byte);
  begin
    for n := 1 to n do
      begin
	getlen(codadr);
	inc(codadr,len)
      end
  end;
  procedure onedown;
  begin
    if keyflags and 16 <> 0
      then down(1)
      else
	if posy[2] = 13
	  then
	    begin
	      down(8);
	      posy[2] := 6
	    end
	  else inc(posy[2])
  end;
  procedure resetw;
  begin
    if keyflags and 3 <> 0
      then { Shift-Esc }
	begin
	  if keyflags and 16 = 0 then posy[3] := 13;
	  stkofs := posy[3]-13
	end
      else { Esc }
	begin
	  codadr := r.pc;
	  if keyflags and 16 = 0 then posy[2] := 6;
	  up(posy[2]-5,true)
	end
  end;

  { Procedures to handle breakpoints }

  procedure delbpt(i: byte);
  begin
    for i := i to 5 do bpt[i] := bpt[i+1];
    with bpt[6] do
      begin
	adr := 0;
	page := 0;
	count := 0;
	occur := 0
      end
  end;
  function fndbpt(a: word; p: byte; del: boolean): byte;
    var i: byte;
  begin
    for i := 1 to 6 do
      with bpt[i] do
	if (count <> 0) and (adr = a) and ((page = 0) or (page = p)) then
	   begin
	     if del then delbpt(i);
	     fndbpt := i;
	     exit
	   end;
    fndbpt := 0
  end;

  { Routine to output a 6800 instruction }

  const mnem: array[byte] of string[4] =
       ('    ','NOP ','    ','    ','    ','    ','TAP ','TPA ',
	'INX ','DEX ','CLV ','SEV ','CLC ','SEC ','CLI ','SEI ',
	'SBA ','CBA ','    ','    ','NBA ','    ','TAB ','TBA ',
	'    ','DAA ','    ','ABA ','    ','    ','    ','    ',
	'BRA ','    ','BHI ','BLS ','BCC ','BCS ','BNE ','BEQ ',
	'BVC ','BVS ','BPL ','BMI ','BGE ','BLT ','BGT ','BLE ',
	'TSX ','INS ','PULA','PULB','DES ','TXS ','PSHA','PSHB',
	'    ','RTS ','    ','RTI ','    ','    ','WAI ','INT ',
	'NEGA','    ','    ','COMA','LSRA','    ','RORA','ASRA',
	'ASLA','ROLA','DECA','    ','INCA','TSTA','    ','CLRA',
	'NEGB','    ','    ','COMB','LSRB','    ','RORB','ASRB',
	'ASLB','ROLB','DECB','    ','INCB','TSTB','    ','CLRB',
	'NEG ','    ','    ','COM ','LSR ','    ','ROR ','ASR ',
	'ASL ','ROL ','DEC ','    ','INC ','TST ','JMP ','CLR ',
	'NEG ','    ','    ','COM ','LSR ','    ','ROR ','ASR ',
	'ASL ','ROL ','DEC ','    ','INC ','TST ','JMP ','CLR ',
	'SUBA','CMPA','SBCA','    ','ANDA','BITA','LDAA','    ',
	'EORA','ADCA','ORAA','ADDA','CPX ','BSR ','LDS ','    ',
	'SUBA','CMPA','SBCA','    ','ANDA','BITA','LDAA','STAA',
	'EORA','ADCA','ORAA','ADDA','CPX ','    ','LDS ','STS ',
	'SUBA','CMPA','SBCA','    ','ANDA','BITA','LDAA','STAA',
	'EORA','ADCA','ORAA','ADDA','CPX ','JSR ','LDS ','STS ',
	'SUBA','CMPA','SBCA','    ','ANDA','BITA','LDAA','STAA',
	'EORA','ADCA','ORAA','ADDA','CPX ','JSR ','LDS ','STS ',
	'SUBB','CMPB','SBCB','    ','ANDB','BITB','LDAB','    ',
	'EORB','ADCB','ORAB','ADDB','    ','    ','LDX ','    ',
	'SUBB','CMPB','SBCB','    ','ANDB','BITB','LDAB','STAB',
	'EORB','ADCB','ORAB','ADDB','    ','    ','LDX ','STX ',
	'SUBB','CMPB','SBCB','    ','ANDB','BITB','LDAB','STAB',
	'EORB','ADCB','ORAB','ADDB','    ','    ','LDX ','STX ',
	'SUBB','CMPB','SBCB','    ','ANDB','BITB','LDAB','STAB',
	'EORB','ADCB','ORAB','ADDB','    ','    ','LDX ','STX ');
  procedure instrout(var w: word);
    label print;
    var i: byte;
	ea, oldattr: word;
  begin
    getlen(w);
    if ypos = posy[2] then
      begin
	attr := attr shl 4 and $7f00;
	cursadr := w
      end;
    if w = r.pc then attr := attr xor $800;
    oldattr := attr;
    if fndbpt(w,newpage,false) <> 0 then attr := attr xor $800;
    wrdcol(w);
    attr := oldattr;
    for i := 1 to len do
      begin
	bytout(get(w));
	inc(w);
	writec(' ')
      end;
    for i := len to 3 do writes('   ');
    if ypos = posy[2] then
      begin
	line := '';
	inc(echo)
      end;
    writes(mnem[opcod]);
    writes('   ');
    case opmod of
      relative:
	begin
	  ea := w+shortint(get(w-1));
	  wrdout(ea)
	end;
      implied:;
      else
	begin
	  if opmod = immed then writec('#');
	  if len = 2
	    then
	      begin
		ea := get(w-1);
		bytout(ea)
	      end
	    else
	      begin
		ea := getw(w-2);
		wrdout(ea)
	      end;
	  if opmod = indexed then
	    begin
	      inc(ea,r.x);
	      writes(',X')
	    end
	end
    end;
    if ypos = posy[2] then
      begin
	case opcod of
	  $22: if not(flags[0] or flags[2]) then goto print;
	  $23: if (flags[0] or flags[2]) then goto print;
	  $24: if not flags[0] then goto print;
	  $25: if flags[0] then goto print;
	  $26: if not flags[2] then goto print;
	  $27: if flags[2] then goto print;
	  $28: if not flags[1] then goto print;
	  $29: if flags[1] then goto print;
	  $2a: if not flags[3] then goto print;
	  $2b: if flags[3] then goto print;
	  $2c: if not(flags[3] xor flags[1]) then goto print;
	  $2d: if flags[3] xor flags[1] then goto print;
	  $2e: if not(flags[2] or (flags[3] xor flags[1])) then goto print;
	  $2f: if flags[2] or (flags[3] xor flags[1]) then goto print;
	  $20, $6e, $7e: print:
	    if ea < cursadr
	      then writes(' '#$18)
	      else
		if ea >= w
		  then writes(' '#$19)
		  else writes(' '#$1b)
	end;
	echo := false
      end;
    while xpos < 39 do writec(' ');
    if (ypos = posy[2]) and (opmod >= direct)
     and (opcod <> $7e) and (opcod <> $bd) then
      begin
	setxy(26,4);
	writec(' ');
	if opmod = direct
	  then bytout(ea)
	  else wrdout(ea);
	if (opcod <> $6e) and (opcod <> $ad) then
	  begin
	    writes(' = ');
	    case opcod and $8f of
	      $8c, $8e, $8f:
		wrdout(getw(ea))
	      else
		if opcod = $3f
		  then
		    begin
		      i := get(intpags+ea shr 1) shr (ord(not odd(ea)) shl 2) and $f;
		      if (i <> 0) and (i <> newpage) then
			begin
			  hexout(i);
			  writec(' ')
			end;
		      wrdout(getw(intadrs+ea shl 1))
		    end
		  else bytout(get(ea))
	    end
	  end;
	writec(' ')
      end
  end;

  { Routine to output data }

  procedure datout(w: word; l: byte);
    var j: byte;
  begin
    wrdcol(w);
    for j := 0 to l-1 do
      begin
	bytout(get(w+j));
	writec(' ')
      end;
    if l = 3
      then writec(' ')
      else
	begin
	  inc(xpos,3);
	  inc(textaddr,6)
	end;
    for j := 0 to l-1 do
      writec(chr(get(w+j)))
  end;

  { Routine to refresh the screen }

  procedure updscr;
    var i: byte;
	w: word;
  begin
    attr := col[1];
    setxy(4,3);
    with r do
      begin
	wrdout(pc); writec(' ');
	hexout(newpage); writes('  ');
	bytout(b); writec(' ');
	bytout(a); writes('  ');
	wrdout(x); writes('  ');
	wrdout(sp); writes('  ')
      end;
    for i := 7 downto 0 do
      if flags[i]
	then writec('1')
	else writec('0');
    attr := swap(lightgray);
    fillx(13,26,4);
    w := codadr;
    for i := 1 to 10 do
      begin
	setxy(4,4+i);
	attr := col[2];
	instrout(w)
      end;
    for i := 12 downto 0 do
      begin
	attr := col[3];
	setxy(44,14-i);
	if stkofs+i = 1 then attr := attr xor $800;
	wrdcol(r.sp+stkofs+i);
	bytout(get(r.sp+stkofs+i))
      end;
    for i := 1 to 6 do
      with bpt[i] do
	begin
	  attr := col[4];
	  setxy(58,1+i);
	  if (count <> 0) and (adr = r.pc) and ((page = 0) or (page = newpage))
	    then attr := attr xor $800;
	  if (count <> 0) and bad then attr := attr xor $800 or $8000;
	  wrdcol(adr);
	  wrdout(count);
	  writes('  ');
	  wrdout(occur);
	  writes('  ');
	  if page <> 0 then hexout(page) else writec(' ');
	  attr := col[5];
	  setxy(58,8+i);
	  datout(wpt[i],3)
	end;
    for i := 0 to 7 do
      begin
	attr := col[6];
	setxy(4,16+i);
	datout(datadr+i shl 4,16)
      end
  end;

  { Routines to save and restore the screen }

  procedure savescr;
  begin
    move(ptr($b800,0)^,scrbuf,sizeof(scrbuf));
    if not model then textmode(co80)
  end;

  procedure out(i,x: byte);
  begin
    port[$3d4] := i;
    port[$3d5] := x
  end;
  procedure outadrs;
  begin
    out(12,lo(crtregs[1]));
    out(13,hi(crtregs[1]))
  end;
  procedure restorescr;
  begin
    if not model then textmode(co40);
    move(scrbuf,ptr($b800,0)^,sizeof(scrbuf));
    regs.cx := swap(crtregs[0]);
    regs.ah := 1;   { Cursor shape setting is done by BIOS to avoid problems }
    intr($10,regs); { with EGA or VGA }
    outadrs;
    out(14,lo(crtregs[2]));
    out(15,hi(crtregs[2]))
  end;

  { Routines to initialize the screen }

  procedure makescr;
    var i, j: byte;
  begin
    regs.ax := $500;
    intr($10,regs);
    regs.ah := 1;
    regs.cx := $d;
    intr($10,regs);
    clrscr; attr := swap(lightgray);
    fillx(78,2,1); fillx(78,2,24); fillx(39,2,4); fillx(78,2,15); fillx(25,55,8);
    filly(22,1,2); filly(23,80,2); filly(13,41,2); filly(13,55,2); filly(8,60,16);
    putc(#$da,1,1); putc(#$cf,80,1); putc(#$c0,1,24); putc(#$d9,80,24);
    putc(#$c2,41,1); putc(#$c2,55,1); putc(#$c2,60,15);
    putc(#$c1,55,15); putc(#$c1,60,24);
    putc(#$c3,1,4); putc(#$c3,1,15); putc(#$c3,55,8);
    putc(#$d4,41,4); putc(#$d4,80,8);
    attr := col[1]; setxy(16,1); writes(' Registers ');
    setxy(5,2); writes('PC  #   B  A    X    SP   11HINZVC');
    attr := col[2]; setxy(11,4); writes(' Code ');
    attr := col[3]; setxy(45,1); writes(' Stack ');
    attr := col[4]; setxy(64,1); writes(' Break ');
    attr := col[5]; setxy(64,8); writes(' Watch ');
    attr := col[6]; setxy(38,15); writes(' Data ');
    setxy(1,25);
    attr := swap(white); writes(' F1     F2        F3      F4       F5       F6       F7        F8       F9 ');
    attr := swap(lightgray);
    writes('Quit'#8#8#8#8#8#8#8#8#8#8#8#8#8'Step'#8#8#8#8#8#8#8#8#8#8#8#8#8#8'Trace'#8#8#8#8#8#8#8#8#8#8#8#8#8#8);
    writes('Next'#8#8#8#8#8#8#8#8#8#8#8#8#8'Look'#8#8#8#8#8#8#8#8#8#8#8#8#8'Here'#8#8#8#8#8#8#8#8#8#8#8#8);
    writes('End'#8#8#8#8#8#8#8#8#8#8#8#8#8'Break'#8#8#8#8#8#8#8#8#8#8#8#8'Go');
    updscr
  end;
  procedure clrs(s: string);
  begin
    clrscr;
    regs.ah := 1;
    regs.cx := oldc;
    intr($10,regs);
    attr := swap(white); setxy(1,1);
    writes(s)
  end;

  { Routines to get keys }

  function getkey: word;
    var c: char;
  begin
    gotoxy(xpos,ypos);
    c := upcase(readkey);
    if c = #0
      then getkey := swap(ord(readkey))
      else getkey := ord(c)
  end;

  procedure setnew; forward;
  procedure setold; forward;
  function fnkeys(key: byte): boolean;
    label there, trap;
    var i: word;
	sstk: longint;
  begin
    fnkeys := true;
    case key of
      $3b: fnkeys := false;
      $3c:
	if fndbpt(cursadr,newpage,true) = 0 then
	  for i := 1 to 6 do
	    with bpt[i] do
	      if count = 0 then
		begin
		  adr := cursadr;
		  if hi(adr) and $e0 = $c0
		    then page := newpage
		    else page := 0;
		  count := 1;
		  occur := 0;
		  bad := false;
		  exit
		end;
      $3d:
	begin
	  endnow := true;
	  oldlvl := r.sp;
	  goto trap
	end;
      $3e:
	begin
	  len := 0;
	  goto there
	end;
      $3f:
	begin
	  restorescr;
	  i := ord(readkey);
	  if i = 0 then i := ord(readkey);
	  savescr;
	  makescr
	end;
      $40:
	begin
	  inc(window);
	  if window = 8 then window := 1
	end;
      $59:
	begin
	  dec(window);
	  if window = 0 then window := 7
	end;
      $41:
	begin
	  trap: r.pc := cursadr;
	  trapnow := true;
	  fnkeys := false
	end;
      $42:
	with bpt[0] do
	  begin
	    r.pc := cursadr;
	    getlen(cursadr);
	    there: adr := cursadr+len;
	    if fndbpt(adr,newpage,false) <> 0 then exit;
	    if hi(adr) and $e0 = $c0
	      then page := newpage
	      else page := 0;
	    count := 1;
	    fnkeys := false
	  end;
      $43, $44:
	begin
	  regs.ah := $4c;
	  msdos(regs)
	end;
      $5c, $5d:
	begin
	  clrs('Type EXIT to return to the  Emulator...');
	  setold;
	  swapvectors;
	  sstk := meml[seg(doserror):ofs(doserror)+2]; { Save/restore SS:SP saved value }
	  exec(getenv('COMSPEC'),'');                  { before and after EXEC to avoid }
	  meml[seg(doserror):ofs(doserror)+2] := sstk; { system crash!!! }
	  swapvectors;
	  setnew;
	  makescr
	end
     end
  end;

  { Routines to get lines }

  function hex(c: word): byte;
  begin
    hex := 16;
    if not bin and (c-ord('0') <= 9) or (c-ord('0') <= 1) then hex := c-ord('0');
    if not bin and (c-ord('A') <= 5) then hex := c-ord('A')+10
  end;

  function hexin(var value: word; i: byte; len: byte): byte;
    label right;
    var key: word;
	j: byte;
  begin
    i := i shl 2;
    repeat
      key := getkey;
      posx[window] := xpos;
      if key = ord(' ') then
	if keyflags and 3 = 0
	  then key := 9     { Convert Space to Tab }
	  else key := $f00; { Convert Shift-Space to Shift-Tab }
      hexin := key;
      if lo(key) <> 0
	then
	  case key of
	    9, $d: exit; { Tab, Enter }
	    $1b: { Esc }
	      begin
		resetw;
		exit
	      end
	    else
	      begin
		j := hex(key);
		if j < 16 then
		  begin
		    writex(chr(key));
		    value := value and not (15 shl i) or j shl i;
		    hexin := $ff;
		    goto right
		  end
	      end
	  end
	else
	  begin
	    hexin := hi(key);
	    case hi(key) of
	      $4b: { Left }
		begin
		  inc(i,4);
		  if i > len then exit;
		  dec(xpos);
		  dec(textaddr,2)
		end;
	      $4d: { Right }
		begin
		  right: if i = 0 then exit;
		  dec(i,4);
		  inc(xpos);
		  inc(textaddr,2)
		end;
	      $3b..$44, $59, $5c, $5d:
		begin
		  hexin := ord(fnkeys(hi(key)));
		  exit
		end;
	      $f, $47..$49, $4f..$51, $53: { Shift-Tab, Home, End, Up, Down, PgUp, PgDn, Del }
		exit
	    end
	  end
    until false
  end;

  function linein(len: byte): byte;
    label start, right;
    var key: word;
	i: byte;
  begin
    for i := length(line) downto 1 do
      if line[i] = ' '
	then dec(line[0])
	else goto start;
    start: for i := 1 to len do
      begin
	if i > length(line)
	  then writex(' ')
	  else writex(line[i]);
	inc(textaddr,2)
      end;
    setxy(xpos, ypos);
    i := 1;
    repeat
      key := getkey;
      if (key = ord(' ')) and (keyflags and 3 <> 0) then key := $f00;
      linein := key;
      if lo(key) <> 0
	then
	  case key of
	    9: { Tab }
	    if i >= 8
	      then exit
	      else
		begin
		  inc(xpos,8-i);
		  inc(textaddr,(8-i) shl 1);
		  i := 8
		end;
	    ord(' '): { Space }
	      if i >= 8
		then
		  begin
		    while i <= len do
		      begin
			line[i] := ' ';
			inc(i)
		      end;
		    linein := $d;
		    exit
		  end
		else
		  while i < 8 do
		    begin
		      line[i] := ' ';
		      writex(' ');
		      inc(xpos);
		      inc(textaddr,2);
		      inc(i)
		    end;
	    $d: exit; { Enter }
	    $1b: { Esc }
	      begin
		line := '';
		dec(xpos,i-1);
		setxy(xpos,ypos);
		goto start
	      end;
	    8: { Backspace }
	      if (i > length(line)) and (length(line) > 0) then
		begin
		  dec(line[0]);
		  dec(xpos);
		  dec(textaddr,2);
		  dec(i);
		  writex(' ')
		end;
	    $20..$7f:
		if i <= len
		  then
		    begin
		      writex(chr(key));
		      while i > length(line) do line := line+' ';
		      line[i] := chr(key);
		      goto right
		    end
		  else writex(' ')
	  end
	else
	  begin
	    linein := hi(key);
	    case hi(key) of
	      $4b: { Left }
		if i > 1
		  then
		    begin
		      dec(xpos);
		      dec(textaddr,2);
		      dec(i)
		    end
		  else exit;
	      $4d: { Right }
		right:
		if i <= len then
		  begin
		    inc(xpos);
		    inc(textaddr,2);
		    inc(i)
		  end;
	      $47: { Home }
		begin
		  dec(xpos,i-1);
		  setxy(xpos,ypos);
		  i := 1
		end;
	      $4f: { End }
		begin
		  inc(xpos,length(line)+1-i);
		  setxy(xpos,ypos);
		  i := length(line)+1
		end;
	      $3b..$44, $59, $5c, $5d:
		begin
		  linein := ord(fnkeys(hi(key)));
		  exit
		end;
	      $f: { Shift-Tab }
	      if i < 8
		then exit
		else
		  begin
		    dec(xpos,i-1);
		    dec(textaddr,i shl 1-2);
		    i := 1
		  end;
	      $48, $49, $50, $51: { Up, Down, PgUp, PgDn }
		exit
	    end
	  end
    until false
  end;

  { Routine to input a 6800 instruction }

  procedure assemble;
    label putit;
    var i, mode: byte;
	val: word;
	s: string[4];
  begin
    i := pos(' ',line); if i = 0 then i := length(line)+1; { Get first space }
    s := copy(line,1,i-1); delete(line,1,i);	 { Move opcode from line to s }
    while pos(' ',line) = 1 do delete(line,1,1); { Remove leading spaces }
    if length(s) = 3 then s := s+' ';            { Blank-pad opcode }
    if length(line) = 0
      then mode := implied
      else
	begin
	  mode := extended;
	  if line[1] = '#' then                  { Check for immediate addressing }
	    begin
	      delete(line,1,1);
	      mode := immed
	    end;
	  val := 0;
	  repeat				 { Get hex value }
	    i := hex(ord(line[1]));
	    if i < 16 then
	      begin
		delete(line,1,1);
		val := val shl 4+i
	      end
	  until (i = 16) or (line = '');
	  if (mode = extended) and (hi(val) = 0) then
	    begin
	      mode := direct;
	      if copy(line,1,2) = ',X'
		then mode := indexed		 { Check for indexed addressing }
	    end
	end;
    for opcod := 0 to 255 do
      if s = mnem[opcod] then
	begin
	  getmod;
	  if (opmod = relative) and ((mode = extended) or (mode = direct))
	    then
	      begin
		dec(val,cursadr+2);
		if shortint(val) = integer(val) then goto putit
	      end
	    else
	      if (mode = opmod) or (mode = direct) and (opmod = extended) then
		begin
		  putit: put(cursadr,opcod);
		  if len = 1 then exit;
		  if len = 2
		    then put(cursadr+1,val)
		    else putw(cursadr+1,val);
		  exit
		end
	end
  end;

  { Main debugging routine }

  procedure debug;
    label right;
    var key, i, j: word;
	k: byte;
    procedure datdn;
    begin
      if (keyflags and 16 <> 0) or (ypos = 23)
	then inc(datadr,16)
	else
	  begin
	    inc(posy[6]);
	    inc(posy[7])
	  end
    end;
    procedure datup;
    begin
      if (keyflags and 16 <> 0) or (ypos = 16)
	then dec(datadr,16)
	else
	  begin
	    dec(posy[6]);
	    dec(posy[7])
	  end
    end;
  begin
    repeat
      attr := col[window];
      setxy(posx[window],posy[window]);
      case window of
	1:
	  begin
	    case xpos of
	      4..7:
		begin
		  key := hexin(r.pc,7-posx[1],12);
		  case key of
		    9, $4d, $ff: posx[1] := 9; { Tab, Right }
		    $f: posx[1] := 31; { Shift-Tab }
		    $4b: posx[1] := 38; { Left }
		    $48: dec(r.pc); { Up }
		    $49: dec(r.pc,16); { PgUp }
		    $50: inc(r.pc); { Down }
		    $51: inc(r.pc,16); { PgDn }
		    $53: r.pc := getw(resvect) { Del }
		  end
		end;
	      9:
		begin
		  key := hexin(newpage,0,0);
		  if newpage < 7 then newpage := 7;
		  case key of
		    9, $4d, $ff: posx[1] := 12; { Tab, Right }
		    $f: posx[1] := 4; { Shift-Tab }
		    $4b: posx[1] := 7; { Left }
		    $48, $49: dec(newpage); { Up, PgUp }
		    $50, $51: inc(newpage); { Down, PgDn }
		    $53: newpage := 7 { Del }
		  end;
		  if newpage = 6 then newpage := $f;
		  if newpage = 16 then newpage := 7
		end;
	      12..13:
		begin
		  key := hexin(r.b,13-posx[1],4);
		  case key of
		    9, $4d, $ff: posx[1] := 15; { Tab, Right }
		    $f, $4b: posx[1] := 9; { Shift-Tab, Left }
		    $48: dec(r.b); { Up }
		    $49: dec(r.b,16); { PgUp }
		    $50: inc(r.b); { Down }
		    $51: inc(r.b,16); { PgDn }
		    $53: r.b := 0 { Del }
		  end
		end;
	      15..16:
		begin
		  key := hexin(r.a,16-posx[1],4);
		  case key of
		    9, $4d, $ff: posx[1] := 19; { Tab, Right }
		    $f: posx[1] := 12; { Shift-Tab }
		    $4b: posx[1] := 13; { Left }
		    $48: dec(r.a); { Up }
		    $49: dec(r.a,16); { PgUp }
		    $50: inc(r.a); { Down }
		    $51: inc(r.a,16); { PgDn }
		    $53: r.a := 0 { Del }
		  end
		end;
	      19..22:
		begin
		  key := hexin(r.x,22-posx[1],12);
		  case key of
		    9, $4d, $ff: posx[1] := 25; { Tab, Right }
		    $f: posx[1] := 15; { Shift-Tab }
		    $4b: posx[1] := 16; { Left }
		    $48: dec(r.x); { Up }
		    $49: dec(r.x,16); { PgUp }
		    $50: inc(r.x); { Down }
		    $51: inc(r.x,16); { PgDn }
		    $53: r.x := 0 { Del }
		  end
		end;
	      25..28:
		begin
		  key := hexin(r.sp,28-posx[1],12);
		  case key of
		    9, $4d, $ff: posx[1] := 31; { Tab, Right }
		    $f: posx[1] := 19; { Shift-Tab }
		    $4b: posx[1] := 22; { Left }
		    $48: dec(r.sp); { Up }
		    $49: dec(r.sp,16); { PgUp }
		    $50: inc(r.sp); { Down }
		    $51: inc(r.sp,16); { PgDn }
		    $53: r.sp := 0 { Del }
		  end
		end;
	      31..38:
		begin
		  i := ord(flags[38-posx[1]]);
		  inc(bin);
		  key := hexin(i,0,0);
		  dec(bin);
		  flags[38-posx[1]] := i <> 0;
		  case key of
		    9: posx[1] := 4; { Tab }
		    $f: posx[1] := 25; { Shift-Tab }
		    $4d, $ff: { Right }
		      begin
			inc(posx[1]);
			if posx[1] = 39 then posx[1] := 4
		      end;
		    $4b: { Left }
		      begin
			dec(posx[1]);
			if posx[1] = 30 then posx[1] := 28
		      end;
		    $48, $49, $50, $51: { Up, PgUp, Down, PgDn }
		      flags[38-posx[1]] := not flags[38-posx[1]];
		    $53: flags[38-posx[1]] := false { Del }
		  end
		end
	    end;
	    case key of
	      0: exit;
	      $47, $d: posx[1] := 4; { Home, Enter }
	      $4f: posx[1] := 38 { End }
	    end
	  end;
	2:
	  begin
	    if xpos = 23
	      then
		begin
		  key := linein(16);
		  assemble;
		  case key of
		    9, $f, $4d: posx[2] := 4; { Tab, Shift-Tab, Right }
		    $4b: posx[2] := 7; { Left }
		    $d: onedown { Enter }
		  end
		end
	      else
		begin
		  i := cursadr;
		  key := hexin(cursadr,7-xpos,12);
		  if (i <> cursadr) and (key <> $1b) then
		    begin
		      codadr := cursadr;
		      up(ypos-5,true)
		    end;
		  case key of
		    9, $f, $4d: posx[2] := 23; { Tab, Shift-Tab, Right }
		    $d, $ff: posx[2] := 4; { Enter }
		    $47: posy[2] := 6; { Home }
		    $4f: posy[2] := 13 { End }
		  end
		end;
	    case key of
	      0: exit;
	      $48: { Up }
		if keyflags and 16 <> 0
		  then up(1,false)
		  else
		    if posy[2] <= 6
		      then
			begin
			  up(8,false);
			  posy[2] := 13
			end
		      else dec(posy[2]);
	      $50: onedown; { Down }
	      $49: up(9,false); { PgUp }
	      $51: down(9) { PgDn }
	    end
	  end;
	3:
	  begin
	    i := r.sp+stkofs+14-ypos;
	    if xpos < 51
	      then
		begin
		  key := hexin(i,47-xpos,12);
		  if key <> $1b then stkofs := i-r.sp+ypos-14;
		  case key of
		    9, $f, $4d: posx[3] := 51; { Tab, Shift-Tab, Right }
		    $ff: posx[3] := 44
		  end
		end
	      else
		begin
		  j := get(i);
		  key := hexin(j,52-xpos,4);
		  if key <> $1b then put(i,j);
		  case key of
		    9, $f: posx[3] := 44; { Tab, Shift-Tab }
		    $4b: { Left }
		      begin
			key := $48;
			posx[3] := 52
		      end;
		    $4d: { Right }
		      begin
			key := $50;
			posx[3] := 51
		      end
		  end
		end;
	    case key of
	      0: exit;
	      $d: posx[3] := 44; { Enter }
	      $47: posy[3] := 2; { Home }
	      $4f: posy[3] := 14; { End }
	      $48: { Up }
		if (keyflags and 16 <> 0) or (posy[3] = 2)
		  then inc(stkofs)
		  else dec(posy[3]);
	      $50: { Down }
		if (keyflags and 16 <> 0) or (posy[3] = 14)
		  then dec(stkofs)
		  else inc(posy[3]);
	      $49: inc(stkofs,13); { PgUp }
	      $51: dec(stkofs,13) { PgDn }
	    end
	  end;
	4:
	  with bpt[ypos-1] do
	    begin
	      if xpos < 65
		then
		  begin
		    i := adr;
		    key := hexin(adr,61-xpos,12);
		    case key of
		      9, $4d, $ff: posx[4] := 65; { Tab, Right }
		      $f: { Shift-Tab }
			if page = 0 then posx[4] := 71 else posx[4] := 77;
		      $4b: { Left }
			if page = 0 then posx[4] := 74 else posx[4] := 77;
		      $49: dec(adr); { PgUp }
		      $51: inc(adr) { PgDn }
		    end;
		    if hi(adr) and $e0 <> $c0
		      then page := 0
		      else
			if page = 0 then page := newpage;
		    if i <> adr then
		      begin
			bad := false;
			occur := 0;
			i := count; count := 0;
			if fndbpt(adr,page,true) <> 0 then;
			if i = 0
			  then count := 1
			  else count := i
		      end;
		  end
		else
		  if xpos < 71
		    then
		      begin
			key := hexin(count,68-xpos,12);
			case key of
			  9, $4d, $ff: posx[4] := 71; { Tab, Right }
			  $f: posx[4] := 58; { Shift-Tab }
			  $4b: posx[4] := 61; { Left }
			  $49: dec(count); { PgUp }
			  $51: inc(count) { PgDn }
			end
		      end
		    else
		      if xpos < 77
			then
			  begin
			    key := hexin(occur,74-xpos,12);
			    case key of
			      9, $4d, $ff: { Tab, Right }
				if page = 0 then posx[4] := 58 else posx[4] := 77;
			      $f: posx[4] := 65; { Shift-Tab }
			      $4b: posx[4] := 68; { Left }
			      $49: dec(occur); { PgUp }
			      $51: inc(occur) { PgDn }
			    end
			  end
			else
			  begin
			    i := page;
			    key := hexin(i,0,0);
			    if i < 7 then i := 7;
			    case key of
			      9, $4d, $ff: posx[4] := 58; { Tab, Right }
			      $f: posx[4] := 71; { Shift-Tab }
			      $4b: posx[4] := 74; { Left }
			      $49: dec(i); { PgUp }
			      $51: inc(i) { PgDn }
			    end;
			    if i = 6 then i := $f;
			    if i = 16 then i := 7;
			    page := i
			  end;
	      case key of
		0: exit;
		$d: posx[4] := 58; { Enter }
		$47: posy[4] := 2; { Home }
		$4f: posy[4] := 7; { End }
		$48: { Up }
		  if posy[4] = 2
		    then posy[4] := 7
		    else dec(posy[4]);
		$50: { Down }
		  if posy[4] = 7
		    then posy[4] := 2
		    else inc(posy[4]);
		$53: delbpt(ypos-1) { Del }
	      end
	    end;
	5:
	  begin
	    if xpos < 65
	      then
		begin
		  key := hexin(wpt[ypos-8],61-xpos,12);
		  case key of
		    9, $4d, $ff: posx[5] := 65; { Tab, Right }
		    $f: posx[5] := 71; { Shift-Tab }
		    $4b: posx[5] := 72; { Left }
		    $49: dec(wpt[ypos-8]); { PgUp }
		    $51: inc(wpt[ypos-8]) { PgDn }
		  end
		end
	      else
		begin
		  i := wpt[ypos-8];
		  k := (xpos-65) div 3;
		  inc(i,k);
		  k := k*3+65;
		  j := get(i);
		  key := hexin(j,1-(xpos-65) mod 3,4);
		  case key of
		    9, $4d, $ff: { Tab, Right }
		      if k = 71
			then posx[5] := 58
			else posx[5] := k+3;
		    $f: { Shift-Tab }
		      if k = 65
			then posx[5] := 58
			else posx[5] := k-3;
		    $4b: { Left }
		      if k = 65
			then posx[5] := 61
			else dec(posx[5],2);
		    $49: dec(j); { PgUp }
		    $51: inc(j) { PgDn }
		  end;
		  put(i,j)
		end;
	    case key of
	      0: exit;
	      $d: posx[5] := 58; { Enter }
	      $47: posy[5] := 9; { Home }
	      $4f: posy[5] := 14; { End }
	      $48: { Up }
		if posy[5] = 9
		  then posy[5] := 14
		  else dec(posy[5]);
	      $50: { Down }
		if posy[5] = 14
		  then posy[5] := 9
		  else inc(posy[5])
	    end
	  end;
	6, 7:
	  begin
	    i := datadr+(ypos-16)*16;
	    if xpos < 11
	      then
		begin
		  key := hexin(i,7-xpos,12);
		  datadr := i-(ypos-16)*16;
		  case key of
		    9, $4d: posx[6] := 11; { Tab, Right }
		    $f: posx[6] := 56; { Shift-Tab }
		    $ff: posx[6] := 4
		  end
		end
	      else
		if xpos < 62
		  then
		    begin
		      k := (xpos-11) div 3;
		      inc(i,k);
		      k := k*3+11;
		      j := get(i);
		      key := hexin(j,1-(xpos-11) mod 3,4);
		      put(i,j);
		      case key of
			$4d, $ff: { Right }
			  if k = 56
			    then
			      begin
				posx[6] := 11;
				datdn
			      end
			    else inc(posx[6],2);
			$4b: { Left }
			  if k = 11
			    then
			      begin
				posx[6] := 57;
				datup
			      end
			    else dec(posx[6],2);
			9: { Tab }
			  if k = 56
			    then posx[6] := 4
			    else posx[6] := k+3;
			$f: { Shift-Tab }
			  if k = 11
			    then posx[6] := 4
			    else posx[6] := k-3
		      end
		    end
		  else
		    begin
		      key := getkey;
		      if lo(key) <> 0
			then
			  case key of
			    $1b: resetw; { Esc }
			    9: goto right; { Tab }
			    $d: { Enter }
			    else
			      begin
				put(i+xpos-62,key);
				key := $ff;
				goto right
			      end
			  end
			else
			  case hi(key) of
			    $3b..$44, $59, $5c, $5d: key := ord(fnkeys(hi(key)));
			    $47..$49, $4f..$51: key := hi(key);
			    $4d: { Right }
			      right:
			      if posx[7] = 77
				then
				  begin
				    posx[7] := 62;
				    datdn
				  end
				else inc(posx[7]);
			    $f, $4b: { Shift-Tab, Left }
			      if posx[7] = 62
				then
				  begin
				    posx[7] := 77;
				    datup
				  end
				else dec(posx[7])
			  end;
		      posx[6] := (posx[7]-62)*3+11;
		      posy[6] := posy[7]
		    end;
	    case key of
	      0: exit;
	      $d: posx[6] := 4; { Enter }
	      $47: posy[6] := 16; { Home }
	      $4f: posy[6] := 23; { End }
	      $48: datup; { Up }
	      $50: datdn; { Down }
	      $49: dec(datadr,128); { PgUp }
	      $51: inc(datadr,128) { PgDn }
	    end;
	    if posx[6] < 11
	      then posx[7] := 62
	      else posx[7] := (posx[6]-11) div 3+62;
	    posy[7] := posy[6]
	  end
      end;
      updscr
    until false
  end;

  { Put/Remove INT 60 at the end of all instruction handlers }

  procedure patch(w: word);
    var i: word;
  begin
     for i := 0 to patchtop do
       memw[emuseg:patchofs[i]] := w
  end;

  { Init the memory }

  procedure init;
    var f: searchrec;
	i: byte;
    procedure readbin(segment, offset, len: word);
      var bin: file;
    begin
       assign(bin,loadpath+f.name);
       reset(bin,1);
       if ioresult = 0 then
	 begin
	   blockread(bin,ptr(segment,offset)^,len);
	   close(bin);
	   inoutres := 0
	 end
    end;
    var y,m,d,w,h,n,s,c: word;
  begin
    fillchar(ptr(mainseg,1)^,$ffff,#0);
    fillchar(ptr(romseg,1)^,$ffff,#0);
    putw(resmagic,$a55a);
    getdate(y,n,d,w);
    gettime(h,m,s,c);
    put(sysdate,c); put(sysdate+1,s);
    put(sysdate+2,m); put(sysdate+3,h);
    put(sysdate+4,d); put(sysdate+5,n);
    putw(sysdate+6,y);
    findfirst(loadpath+'BIOS601?.BIN',readonly,f);
    if doserror = 0
      then readbin(mainseg,$f000,4096)
      else put(modelbyte,128);
    findfirst(loadpath+'PAGEROM?.BIN',readonly,f);
    while doserror = 0 do
      begin
	i := hex(ord(f.name[8]));
	if (i >= 8) and (i < 16) then readbin(romseg,i shl 13,8192);
	findnext(f)
      end
  end;

  { Break interrupt handler }

  var savseg: word;
  procedure break(fl, cs, ip, ax, bx, cx, dx, si, di, ds, es, bp: word);
    interrupt;
    label enter;
    var i, j: word;
  begin
    if emuseg = 0 then { Was called at startup }
      begin
	emuseg := ax;
	mainseg := bx;
	ramseg := cx;
	romseg := dx;
	regs.ah := 3;
	regs.bh := 0;
	intr($10,regs);
	oldc := regs.cx;
	crtregs[0] := $706;
	crtregs[1] := 0;
	crtregs[2] := 0;
	i := 0;
	j := 0;
	repeat
	  if memw[emuseg:i] = $e1ff then
	    begin
	      patchofs[j] := i;
	      if hi(i) = $39 then patchrts := i;
	      if hi(i) = $3b then patchrti := i;
	      inc(j);
	      if memw[emuseg:i+2] = 0 then i := i and $ff00 or $ff
	    end;
	  inc(i)
	until i = 0;
	patchtop := j-1;
	if hotkey <> 0 then patch(swap(trap)+$cd);
	init;
	model := get(modelbyte) >= 128;
	if model then i := co80 else i := co40;
	if lastmode <> i then textmode(i) else clrscr;
	ax := seg(scrbuf) + ofs(scrbuf) shr 4;
	exit
      end;
    oldpage := mem[mainseg:rompage] and $f;
    if oldpage < 7 then oldpage := 7;
    if skipbpt <> $ff then
      with bpt[skipbpt] do
	begin
	  old := getp(adr,page);
	  putp(adr,brk,page);
	  patch($e1ff);
	  skipbpt := $ff;
	  ip := swap(getp(si-1,oldpage));
	  exit
	end;
    if (hotkey = 0) and not trapnow and not irqnow
      then { Was called by the BRK instruction }
	begin
	  i := 0;
	  while i < 6 do
	    with bpt[i] do
	      if (count <> 0) and (adr = si-1) and ((page = 0) or (page = oldpage))
		then
		  begin
		    inc(occur);
		    if (occur = count) and ((i = 0) or (bpt[0].count = 0))
		      then goto enter
		      else
			begin
			  skipbpt := i;
			  putp(adr,old,page);
			  patch(swap(trap)+$cd);
			  ip := swap(old);
			  exit
			end
		  end
		else inc(i)
	end
      else
	with r do
	  begin
	    patch($e1ff);
	    if (hotkey = $4c) or (hotkey = $54) then
	      begin
		hotkey := 0;
		si := getw(resvect)+1;
		ip := swap(get(si-1));
		mem[emuseg:$ffff] := $d0;
		exit
	      end;
	    if irqnow then { Was called from timer interrupt }
	      begin
		if keytim <> 0 then
		  begin
		    dec(keytim);
		    put(oldkey,$ff) { Bad patch!!! }
		  end;
		regs.ah := 1;
		intr($16,regs); inline($fa); { CLI }
		if regs.flags and fzero = 0
		  then
		    begin
		      regs.ah := 0;
		      intr($16,regs); inline($fa); { CLI }
		      i := regs.al;
		      if (i = 0) and (regs.ah <= $8a) and (regs.ah >= $3b)
			then i := xlat[regs.ah];
		      if regs.ax = $f00 then i := $ee;
		      if i = 8 then i := $7f;
		      if i = 9 then i := $ed;
		      if i = $d then i := $c0;
		      if i <> 0 then
			begin
			  keytim := 1;
			  put(piadra,i);
			  put(piacra,128);
			  put(piacrb,0)
			end
		    end
		  else
		    begin
		      put(piacrb,128);
		      put(piacra,0)
		    end;
		putwp(bp-1,si-1,7);
		putwp(bp-3,di,7);
		putwp(bp-5,swap(dx),7);
		i := mem[emuseg:$ffff];
		putp(bp-6,hi(ax) or fl shr 4 and $c or fl shr 10 and 2 or i,7);
		mem[emuseg:$ffff] := i or $10;
		dec(bp,7);
		si := getw(irqvect)+1;
		ip := swap(get(si-1));
		dec(irqnow);
		exit
	      end;
	    enter: inc(debugnow);
	    inline($fb); { STI }
	    pc := si-1; sp := bp; x := di; a := hi(dx); b := lo(dx);
	    byte(flags[0]) := hi(ax); byte(flags[1]) := fl shr 11 and 1;
	    byte(flags[2]) := fl shr 6 and 1; byte(flags[3]) := fl shr 7 and 1;
	    i := mem[emuseg:$ffff]; byte(flags[4]) := i shr 4 and 1;
	    byte(flags[5]) := i shr 5 and 1; byte(flags[6]) := i shr 6 and 1;
	    byte(flags[7]) := i shr 7;
	    if not endnow or (sp > oldlvl) or (hotkey <> 0) then
	      begin
		hotkey := 0;
		endnow := false;
		if oldpage < 8 then
		  move(ptr(mainseg,$c000)^,ptr(ramseg,0)^,$2000);
		if trapnow
		  then es := savseg
		  else
		    begin
		      for i := 0 to 6 do
			with bpt[i] do
			  if count <> 0 then
			    if getp(adr,page) = brk
			      then
				begin
				  putp(adr,old,page);
				  bad := false
				end
			      else bad := true;
		      bpt[0].count := 0
		    end;
		if (pc <> cursadr) or (hi(pc) and $e0 = $c0) and (newpage <> oldpage) then
		  begin
		    getlen(cursadr);
		    if (pc = cursadr+len) and ((hi(pc) and $e0 <> $c0) or (newpage = oldpage))
		      then onedown
		      else
			begin
			  keyflags := keyflags and $fc;
			  resetw
			end
		  end;
		newpage := oldpage;
		if trapnow
		  then
		    begin
		      dec(trapnow);
		      updscr
		    end
		  else
		    begin
		      savescr;
		      makescr
		    end;
		debug;
		if trapnow
		  then
		    begin
		      if not endnow then patch(swap(trap)+$cd);
		      savseg := es; es := seg(scrbuf) + ofs(scrbuf) shr 4
		    end
		  else
		    begin
		      restorescr;
		      for i := 0 to 6 do
			with bpt[i] do
			  if count <> 0 then
			    begin
			      old := getp(adr,page);
			      putp(adr,brk,page);
			      if occur >= count then occur := 0;
			      if (adr = pc) and ((page = 0) or (page = newpage))
				then dec(occur)
			    end;
		    end;
		if newpage < 8
		  then move(ptr(ramseg,0)^,ptr(mainseg,$c000)^,$2000)
		  else move(ptr(romseg,newpage shl 13)^,ptr(mainseg,$c000)^,$2000);
		put(rompage,newpage)
	      end;
	    if endnow then
	      begin
		memw[emuseg:patchrts] := swap(trap)+$cd;
		memw[emuseg:patchrti] := swap(trap)+$cd
	      end;
	    si := pc+1; ip := swap(get(pc)); bp := sp; di := x;
	    dx := swap(a) or b;
	    ax := swap(ord(flags[0])); fl := $200 or ord(flags[1]) shl 11
	     or ord(flags[2]) shl 6 or ord(flags[3]) shl 7;
	    mem[emuseg:$ffff] := ord(flags[4]) shl 4 or ord(flags[5]) shl 5 or
	     ord(flags[6]) shl 6 or ord(flags[7]) shl 7;
	    dec(debugnow)
	  end
  end;

  { Keyboard interrupt handler }

  procedure keyboard; interrupt;
  begin
    if (not irqnow and not trapnow or endnow) and (skipbpt = $ff) and (hotkey = 0) then
      begin
	hotkey := port[$60];
	if ((hotkey = $38) or (hotkey = $4c) or (hotkey = $54)) and (emuseg <> 0)
	  then patch(swap(trap)+$cd)
	  else hotkey := 0
      end;
    inline($9c);     { PUSHF }
    oldkbd
  end;

  { Timer interrupt handler }

  procedure timer; interrupt;
    var i, j, k: word;
	regs: registers;
  begin
    inc(timecnt);
    if timecnt = 3
      then
	begin
	  inline($9c);	   { PUSHF }
	  oldtim;
	  timecnt := 0
	end
      else port[$20] := $20; { Signal end of interrupt }
    if not interrupts
      then
	begin
	  i := getw(synchroadr); { First set screen base address (for scrolling) }
	  if i <> 0 then
	    begin
	      putw(synchroadr,0);
	      inc(i,$1000);
	      crtregs[1] := swap(i);
	      if not trapnow then outadrs
	    end;
	  inc(mem[mainseg:sysdate],2); { Then increment system time (and date) }
	  if get(sysdate) >= 100 then
	    begin
	      put(sysdate,0);
	      inc(mem[mainseg:sysdate+1]);
	      if get(sysdate+1) >= 60 then
		begin
		  put(sysdate+1,0);
		  inc(mem[mainseg:sysdate+2]);
		  if get(sysdate+2) >= 60 then
		    begin
		      put(sysdate+2,0);
		      inc(mem[mainseg:sysdate+3]);
		      if get(sysdate+3) >= 24 then
			put(sysdate+3,0) { Should increment here date, month,
					   year, but who really cares? }
		    end
		end
	    end;
	  regs.ah := 1; { Finally check for a pressed key and put it into buffer }
	  intr($16,regs); inline($fa); { CLI }
	  if regs.flags and fzero = 0 then
	    begin
	      regs.ah := 0;
	      intr($16,regs); inline($fa); { CLI }
	      i := regs.al;
	      if (i = 0) and (regs.ah <= $8a) and (regs.ah >= $3b)
		then i := xlat[regs.ah];
	      if regs.ax = $f00 then i := $ee;
	      if i = 8 then i := $7f;
	      if i = 9 then i := $ed;
	      if i = $d then i := $c0;
	      if i <> 0 then
		begin
		  j := getw(kbdstore);
		  if not debugnow then
		    begin
		      k := j+1;
		      if k = $c000 then k := $bff0;
		      if k <> getw(kbdload) then
			begin
			  put(j,i);
			  putw(kbdstore,k)
			end
		    end
		end
	    end
	end
      else
	if not debugnow and not irqnow and not trapnow and (hotkey = 0) and
	 (skipbpt = $ff) and (emuseg <> 0) and (mem[emuseg:$ffff] and $10 = 0) then
	  begin
	    patch(swap(trap)+$cd);
	    inc(irqnow)
	  end
  end;

  { Grab interrupt vectors }

  procedure setnew;
  begin
    getintvec(trap,@oldusr);
    setintvec(trap,@break);
    getintvec(9,@oldkbd);
    setintvec(9,@keyboard);
    getintvec(8,@oldtim);
    setintvec(8,@timer);
    port[$40] := $55; { Set timer interrupt to 54.6 Hz (should be 50 Hz) }
    port[$40] := $55
  end;

  { Restore interrupt vectors }

  procedure setold;
  begin
    port[$40] := 0; { Set timer interrupt to 18.2 Hz }
    port[$40] := 0;
    setintvec(8,@oldtim);
    setintvec(9,@oldkbd);
    setintvec(trap,@oldusr)
  end;

  { Routine to initialize variables and start the emulator }

  const nofil: string[24] = 'Can''t read PYLDIN.DAT'#$d#$a'$';
	nomem: string[20] = 'Not enough memory'#$d#$a'$';
  var i, j: byte;
      f: file of byte;
begin
  if (ofs(scrbuf) and 15 <> 0) or (ofs(doserror)+6 <> ofs(checkbreak)) then exit;
  dec(checkbreak);
  fsplit(paramstr(0),loadpath,line,line);
  filemode := 0;
  line := paramstr(1);
  if (length(line) = 2) and (line[1] = '/') and (upcase(line[2]) = 'E') then
    inc(interrupts);
  assign(f,loadpath+'COLORS.CFG');
  reset(f);
  for i := 1 to 6 do
    begin
      j := 0;
      read(f,j);
      case j of
	0, 8..255: j := 7
      end;
      col[i] := swap(j);
    end;
  col[7] := col[6];
  close(f);
  for i := 0 to 6 do
    with bpt[i] do
      begin
	adr := 0;
	page := 0;
	occur := 0;
	count := 0;
	wpt[i] := 0
      end;
  inoutres := 0;
  setnew;
  exec(loadpath+'PYLDIN.DAT','');
  setold;
  if doserror <> 0
    then
      begin
	regs.ds := seg(nofil);
	if doserror = 8
	  then regs.dx := ofs(nomem)+1
	  else regs.dx := ofs(nofil)+1;
	regs.ah := 9;
	msdos(regs);
	halt(255)
      end
    else clrs('Goodbye from the  Emulator (C) 1989-91 Todor Todorov');
  halt(dosexitcode)
end.
